///////////////////////////////////////////////////////
//            NFA
/////////////////////////////////////////////////////
import java.util.*;
import java.io.*;

public class NFA{
    static int Nmax=2285760;  //an upper bound for the nb of states
    static int epsilon = -1;
    Map<Integer, Set<Integer>>[] next;
    Set<Integer> initial;
    Set<Integer> terminal;
    int nbStates;
    /**
       Creates an NFA with <code>n</code> states.
    */
    public NFA(int n){
	next = new TreeMap[n];
	nbStates = 0;
	initial = new TreeSet<Integer>();
	terminal = new TreeSet<Integer>();
    }
    /**
       Creates an NFA with <code>n</code> states 
       on the alphabet <code>a</code>.
    */
    public void SetInitial(int i) {
	initial.add(i);
    }
    public void SetTerminal(int i) {
	terminal.add(i);
    }
    public int newState() {
	int n = nbStates;
	next[n] = new TreeMap<Integer,Set<Integer>>();
	nbStates++;
	return n;
    }
    
    public String toString(){
	String s = "nbStates = "+nbStates+"\n";
	s += "initial = "+initial+"\n";
	s +="terminal = "+terminal+"\n";
	for(int i = 0; i<nbStates; i++){
	    s += i + " : " + next[i]+"\n";
	}
	return s;
    }

    /**
       Returns the set of states which can be reached 
       from <code>p</code>
       by an epsilon path. Implements a depth-first search of the
       graph of epsilon transitions. 
       Uses a boolean array <code>mark</code> 
       for the exploration.
       @param p a state (i.e. a unary half-edge)
       @param mark the array of marks
       @return the set of states obtained.
    */
    public Set<Integer> closure(Integer p, boolean[] mark){
	Set<Integer> res = new TreeSet<Integer>() ;
	res.add(p);
	mark[p] = true;
	if (next[p].containsKey(epsilon) ) {
	    for (Integer e : next[p].get(epsilon) ) {
		if(!mark[e]) {
		    res.addAll(closure(e, mark));
		}
		else res.add(e);
	    }
	}
	return res;
    }
    /**
       Implements the closure of the set of states <code>s</code>.
    */
    public Set<Integer> closure(Set<Integer> s){
	Set<Integer> t = new TreeSet<Integer>();
	boolean[] mark = new boolean[nbStates];
	for (Integer i : s) {
	    t.addAll(closure(i, mark));
	}
	return t;
    }

    /**
       Returns the set of states which can be reached 
       from <code>p</code>
       by an epsilon path. Implements a depth-first search of the
       graph of epsilon transitions. 
       @param p a state (i.e. a unary half-edge)
       @return the set of states obtained.
    */
    public Set<Integer> closure2(Integer p){
	Set<Integer> s = new TreeSet<Integer>();
	s.add(p);
	return closure2(s);
    }
    
    /**
       Implements the closure of the set of states <code>s</code>.
    */
    public Set<Integer> closure2(Set<Integer> s){
	Set<Integer> res = new TreeSet<Integer>();
	Queue<Integer> Q = new LinkedList<Integer>();
	for (Integer e : s)
	    Q.add(e);
	while (!Q.isEmpty()) {
	    Integer p = Q.remove();
	    if (!res.contains(p)) {
		res.add(p);
		if (next[p].containsKey(epsilon) ) {
		    for (Integer e : next[p].get(epsilon) ) {
			Q.add(e);
		    }
		}
	    }
	}
	return res;
    }

    public void addEdge(int p, int a, int q) {
	if (next[p].containsKey(a)) 
	    next[p].get(a).add(q);
	else {
	    Set<Integer> s = new TreeSet<Integer>();
	    s.add(q);
	    next[p].put(a, s);
	}
    }
    /**
       Computes a set transition in a literal NFA.
       @param s the original set of states
       @param c a letter
       @return the set of states reachable from s under input c.
    */
    public Set<Integer> next(Set<Integer> s, int c){
	Set<Integer> t = new TreeSet<Integer>();
	for(Integer e : s){
	    t.addAll(next[e].get(c));
	}
	return closure2(t);
    }

    /*
     * Comparator for Sets represented as Set<Integer>
     */
    private static final Comparator<Set<Integer>> 
	keyComparator1 = new Comparator<Set<Integer>>() {
	    public int compare(Set<Integer> as, Set<Integer> bs) {
		int result;
		if ((result = as.size() - bs.size()) != 0) { 
		    return result;
		}
		Iterator<Integer> i = as.iterator();
		Iterator<Integer> j = bs.iterator();
		for ( ; result == 0 && i.hasNext();) {
		    result =  i.next() - j.next();
		}
		return result;
	    }
	    public boolean equals(Set<Integer> as, Set<Integer> bs) {
		return compare(as, bs) == 0;
	    }

	};

    /*
     * toDFA :
     *  Use a Map<Set<Integer>, Integer> 
     */

    public DFA toDFA(){
	DFA b = new DFA(Nmax);
	Map<Set<Integer>, Integer> t = 
	    new TreeMap<Set<Integer>, Integer>(keyComparator1);
	Queue<Set<Integer>> Q = new LinkedList<Set<Integer>>();
	Set<Integer> I = closure2(initial);
	b.initial = b.newState();
	t.put(I, b.initial);    
	Q.add(I);
	while(! Q.isEmpty()) {
	    Set<Integer> s = Q.remove();
	    Map<Integer, Set<Integer>> sm = new TreeMap<Integer, Set<Integer>>();
	    for (Integer p : s) {
		for (Integer c : next[p].keySet() ) {
		    if (!sm.containsKey(c)) {
			sm.put(c, new TreeSet<Integer>(next[p].get(c)));
		    }
		    else
			sm.get(c).addAll(next[p].get(c)); //merge ?
		}
	    }
	    Integer q = t.get(s);
	    for (Integer c : sm.keySet() ) {
		Set<Integer> sc = closure2(sm.get(c));
		Integer nn = t.get(sc);
		if (nn == null) {
		    int n = b.newState();
		    t.put(sc, n);
		    Q.add(sc);
		    b.addEdge(q,c,n);
		}
		else {
		    b.addEdge(q, c, nn);
		}
	    }
	}
	b.terminal = new TreeSet<Integer>();
	for(Set<Integer> p : t.keySet()) {
	    if (! Collections.disjoint(terminal, new TreeSet<Integer>(p))) {
		b.SetTerminal(t.get(p));
	    }
	}
	return b;
    }

    /*
     * Comparator for Sets represented as Integer[]
     */
    private static final Comparator<Integer[]> 
	keyComparator = new Comparator<Integer[]>() {
	    public int compare(Integer[] as, Integer[] bs) {
		int result;
		if ((result = as.length - bs.length) != 0) { 
		    return result;
		}
		int length = as.length;
		int i = 0;
		while (i < length && (result =as[i] -bs[i]) == 0) {
		    i++;
		}
		return result;
	    }
	    public boolean equals(Integer[] as, Integer[] bs) {
		return compare(as, bs) == 0;
	    }
	};


    /*
     * toDFA :
     *  Use a Map<Integer[], Integer> 
     */

    public DFA toDFA1(){
	DFA b = new DFA(Nmax);
	Map<Integer[], Integer> t = new TreeMap<Integer[], Integer>(keyComparator);
	Queue<Integer[]> Q = new LinkedList<Integer[]>();
	Set<Integer> si  = closure2(initial);  
	Integer[] I = new Integer[si.size()];
	si.toArray(I);
	//	System.out.println(si + " = " +I);
	b.initial = b.newState();
	t.put(I, b.initial);             /* add I to t */
	Q.add(I);
	while(! Q.isEmpty()) {
	    Integer[] s = Q.remove();
	    Map<Integer, Set<Integer>> sm = new TreeMap<Integer, Set<Integer>>();
	    for (Integer p : s) {
		for (Integer c : next[p].keySet() ) {
		    if (!sm.containsKey(c)) {
			sm.put(c, new TreeSet<Integer>(next[p].get(c)));
		    }
		    else
			sm.get(c).addAll(next[p].get(c)); //merge ?
		}
	    }
	    Integer q = t.get(s);
	    for (Integer c : sm.keySet() ) {
		Set<Integer> sc = closure2(sm.get(c));
		Integer[] sci =  new Integer[sc.size()];
		sc.toArray(sci);
		Integer nn = t.get(sci);
		if (nn == null) {
		    int n = b.newState();
		    t.put(sci, n);
		    Q.add(sci);
		    b.addEdge(q,c,n);
		}
		else {
		    b.addEdge(q, c, nn);
		}
	    }
	}
	b.terminal = new TreeSet<Integer>();
	for(Integer[] p : t.keySet()) {
	    if (! Collections.disjoint(terminal, Arrays.asList(p))) {
		b.SetTerminal(t.get(p));
	    }
	}
	return b;
    }
    
        /*
     * toDFA :
     *  Use a Map represented as a TST (ternary search trie)
     */

    public DFA toDFA2(){
	DFA b = new DFA(Nmax);
	TST t = new TST();
	Queue<Integer[]> Q = new LinkedList<Integer[]>();
	Set<Integer> si  = closure2(initial);  
	Integer[] I = new Integer[si.size()];
	si.toArray(I);
	//	System.out.println(si + " = "  + Arrays.asList(I));
	b.initial = b.newState();
	t.put(I, b.initial);             /* add I to t */
	//	System.out.println("TST : " +t);
	
	Q.add(I);
	while(! Q.isEmpty()) {
	    Integer[] s = Q.remove();
	    Map<Integer, Set<Integer>> sm = new TreeMap<Integer, Set<Integer>>();
	    for (Integer p : s) {
		for (Integer c : next[p].keySet() ) {
		    if (!sm.containsKey(c)) {
			sm.put(c, new TreeSet<Integer>(next[p].get(c)));
		    }
		    else
			sm.get(c).addAll(next[p].get(c)); //merge ?
		}
	    }
	    Integer q = t.get(s);
	    for (Integer c : sm.keySet() ) {
		Set<Integer> sc = closure2(sm.get(c));
		Integer[] sci =  new Integer[sc.size()];
		sc.toArray(sci);
		int nn = t.get(sci);
		if (nn == -1) {
		    int n = b.newState();
		    t.put(sci, n);
		    Q.add(sci);
		    b.addEdge(q,c,n);
		}
		else {
		    b.addEdge(q, c, nn);
		}
	    }
	}

	b.terminal = new TreeSet<Integer>();
	for(Integer[] p : t.keySet()) {
	    if (! Collections.disjoint(terminal, Arrays.asList(p))) {
		b.SetTerminal(t.get(p));
	    }
	}
	return b;
    }

    void addWord(String w) {
	if (initial.isEmpty()) {
	    int n = newState();
	    initial.add(n);
	}
	
	int i = 0;
	Integer p = initial.iterator().next();
	while (i < w.length()) {
	    int c = (int) w.charAt(i);
	    if (!next[p].containsKey(c)) {
		int n = newState();
		addEdge(p, c, n);
		p = n;
	    }
	    else {
		p = next[p].get(c).iterator().next();
	    }
	    i++;
	}
	SetTerminal(p);
    }
	
    NFA reverse() {
	NFA a = new NFA(nbStates);
	for (int i = 0; i < nbStates; i++) {
	    a.newState();
	}
	for (int i=0; i < nbStates; i++) {
	    for (Integer c : next[i].keySet()) {
		for (Integer q : next[i].get(c)) 
		    a.addEdge(q, c, i);
	    }
	}
	a.terminal = new TreeSet<Integer>(initial);
	a.initial = new TreeSet<Integer>(terminal);
	return a;
    }
    public static NFA fromFile(String name) throws IOException {
	FileReader fileIn = new FileReader(name);
	BufferedReader r = new BufferedReader(fileIn);
	String line;	
	int count=0;
	NFA a =new NFA(Nmax);
	while ((line = r.readLine()) != null) {
  	    count++;
  	    if(count % 10000 == 0){		
	    System.out.println("nbWords = " + count);
	    }
	    a.addWord(line);
	    }
	fileIn.close();
	System.out.println("Total  = " + count);
	
	return a;
    }
}
